unit exbitcnt;

interface

uses Forms,z_prof;

var
	ans:integer;

implementation

function BitCount0(x:cardinal):integer;
var i:integer;
begin
	result:=0;
	for i:=0 to 31 do
	begin
		inc(result,(x and $1));
		x:=x div 2;
	end;
end;

//* +++Date last modified: 05-Jul-1997 */

//*
//**  Bit counter by Ratko Tomic
//*/

//#include "bitops.h"

//int CDECL bit_count(long x)
{
				int n = 0;
/*
** The loop will execute once for each bit of x set, this is in average
** twice as fast as the shift/test method.
*/
				if (x) do
							n++;
				while (0 != (x = x&(x-1))) ;
				return(n);
}

function BitCount1(x:integer):integer;
begin
	result:=0;
	while x<>0 do
	begin
		inc(result);
		x:=x and (x-1);
	end;
end;

type
	Tbit=0..31;
	TBits=set of TBit;

function BitCount2(x:cardinal):integer;
var
	i:integer;
	s:TBits;

begin
	s:=TBits(x);
	result:=0;
	for i:=0 to 31 do
		if i in s then
			inc(result);
end;

//
// Dense bit count from Dave Shapiro (who got it from some where else
// ~ 3x faster than BitCount1
//

function BitCount3(X: integer): Integer;
begin
  X := X and $55555555 + X shr 1 and $55555555;
  X := X and $33333333 + X shr 2 and $33333333;
  X := X and $0f0f0f0f + X shr 4 and $0f0f0f0f;
  X := X and $00ff00ff + X shr 8 and $00ff00ff;
  X := X and $0000ffff + X shr 16 and $0000ffff;
  Result := X;
end;

// fewer instructions but not noticiablly faster

function BitCount4(X: integer): Integer;
var
  Y:Integer;
begin
  Y:=X;
  X := X and $55555555 + Y shr 1 and $55555555;
  Y:=X;
  X := X and $33333333 + Y shr 2 and $33333333;
  Y:=X;
  X := X and $0f0f0f0f + Y shr 4 and $0f0f0f0f;
  Y:=X;
  X := X and $00ff00ff + Y shr 8 and $00ff00ff;
  Y:=X;
  Result := X and $0000ffff + Y shr 16 and $0000ffff;
end;


// Good ole table lookup.
// Kind of boring but still the winner (excluding table creation) for avg bit counts>~4

var
  BitTable:array[0..255] of integer;

procedure LoadTable;
var i:integer;
begin
  i:=0;
  while i<256 do
//  for i:=0 to 255 do
  begin
    BitTable[i]:=BitCount1(i);
    inc(i);
  end;
end;

function BitCount5(X: integer): Integer;
begin
  Result:=BitTable[X and $FF]+BitTable[(X shr 8) and $FF]+BitTable[(X shr 16) and $FF]+BitTable[(X shr 24)];
end;

function BitCount6(X:integer):integer;
asm
   test eax,eax
   jz @@Done

   push ebx
   mov ebx,eax

   xor ecx,ecx
   xor edx,edx

   mov cl,bl
   mov dl,bh

   shr ebx,16
   mov eax, dword ptr [BitTable+ecx*4]

   add eax, dword ptr [BitTable+edx*4]
   mov cl,bl

   add eax, dword ptr [BitTable+ecx*4]
   mov dl,bh

   add eax, dword ptr [BitTable+edx*4]
   pop ebx
@@Done:
end;

//
// Super_Dense bit count from Charles Doumar (who got it from some where else)
// ~ 3.4 x faster than BitCount1 and at least 20% faster than Bitcount4
//

function BitCount7(X: integer): Integer;
var
  Y:integer;
begin
  Y := X;
  Y := Y - ((Y shr 1) and $55555555);
  Y := (Y and $33333333) + ((Y shr 2) and $33333333);
  Y := Y + Y shr 4;
  Y := Y and $0f0f0f0f;
  Y := Y + Y shr 8;
  y := Y + Y shr 16;
  Result := Y and $000000ff;
end;

Function BitCount8(X: integer): integer;
const
  nibbleTable:array[0..15] of integer=( 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4);
begin
  Result:=BitTable[X and $F];
  Result:=Result+BitTable[(X shr 4) and $F];
  Result:=Result+BitTable[(X shr 8) and $F];
  Result:=Result+BitTable[(X shr 12) and $F];
  Result:=Result+BitTable[(X shr 16) and $F];
  Result:=Result+BitTable[(X shr 20) and $F];
  Result:=Result+BitTable[(X shr 24) and $F];
  Result:=Result+BitTable[(X shr 28)];
end;

procedure RunTests;
var
	test:integer;
  i:Integer;
begin
  //warmup cache
	ans:=BitCount0(test);
	ans:=BitCount1(test);
	ans:=BitCount2(test);
	ans:=BitCount3(test);
	ans:=BitCount4(test);
	ans:=BitCount5(test);
	ans:=BitCount6(test);
  ans:=BitCount7(test);
  ans:=BitCount8(test);
  for i:=0 to 99 do
  begin
    Test:=random(255) shl 24 + random(255) shl 16 + random(255) shl 8 + random(255);
    profile.mark(1,true);
    BitCount0(test);
    profile.mark(1,false);
    profile.mark(2,true);
    BitCount1(test);
    profile.mark(2,false);
    profile.mark(3,true);
    BitCount2(test);
    profile.mark(3,false);
    profile.mark(4,true);
    BitCount3(test);
    profile.mark(4,false);
    profile.mark(5,true);
    BitCount4(test);
    profile.mark(5,false);
    profile.mark(6,true);
    BitCount5(test);
    profile.mark(6,false);
    profile.mark(7,true);
    BitCount6(test);
    profile.mark(7,false);
    profile.mark(8,true);
    BitCount7(test);
    profile.mark(8,false);
    profile.mark(9,true);
    BitCount8(test);
    profile.mark(9,false);
  end;
end;


initialization
  LoadTable;
  Randomize;
	TZProfiler.Create(Application);
  RunTests;
end.
